@@ -5,8 +5,6 @@ from trytond.model import (ModelView, ModelSQL, MatchMixin, fields,
 from trytond.pyson import If, Eval, Bool
 from trytond.pool import PoolMeta
 
-__all__ = ['ProductLocation', 'Move', 'ShipmentIn', 'ShipmentOutReturn']
-
 
 class ProductLocation(sequence_ordered(), ModelSQL, ModelView, MatchMixin):
     '''
@@ -14,8 +12,23 @@ class ProductLocation(sequence_ordered(), ModelSQL, ModelView, MatchMixin):
     It defines the default storage location by warehouse for a product.
     '''
     __name__ = 'stock.product.location'
-    product = fields.Many2One('product.product', 'Product', required=True,
-        select=True, ondelete='CASCADE')
+    template = fields.Many2One(
+        'product.template', "Product",
+        required=True, ondelete='CASCADE', select=True,
+        domain=[
+            If(Bool(Eval('product')),
+                ('products', '=', Eval('product')),
+                ()),
+            ],
+        depends=['product'])
+    product = fields.Many2One(
+        'product.product', "Variant", ondelete='CASCADE', select=True,
+        domain=[
+            If(Bool(Eval('template')),
+                ('template', '=', Eval('template')),
+                ()),
+            ],
+        depends=['template'])
     warehouse = fields.Many2One('stock.location', 'Warehouse', required=True,
         domain=[('type', '=', 'warehouse')], ondelete='CASCADE')
     location = fields.Many2One('stock.location', 'Storage Location',
@@ -26,6 +39,20 @@ class ProductLocation(sequence_ordered(), ModelSQL, ModelView, MatchMixin):
                     [Eval('warehouse')], [])),
             ], depends=['warehouse'])
 
+    @classmethod
+    def __register__(cls, module_name):
+        table = cls.__table_handler__(module_name)
+
+        super().__register__(module_name)
+
+        # Migration from 5.6: Add template on locations
+        table.not_null_action('product', 'remove')
+
+    @fields.depends('product', '_parent_product.template')
+    def on_change_product(self):
+        if self.product:
+            self.template = self.product.template
+
 
 class Move(metaclass=PoolMeta):
     __name__ = 'stock.move'
@@ -35,7 +62,12 @@ class Move(metaclass=PoolMeta):
         if getattr(self, 'shipment', None):
             pattern.setdefault('warehouse', self.shipment.warehouse.id)
 
-        for product_location in self.product.locations:
+        if self.product:
+            pattern.setdefault('template', self.product.template.id)
+            pattern.setdefault('product', self.product.id)
+
+        locations = self.product.locations + self.product.template.locations
+        for product_location in locations:
             if product_location.match(pattern):
                 setattr(self, field, product_location.location)
                 break
@@ -44,23 +76,18 @@ class Move(metaclass=PoolMeta):
 class ShipmentIn(metaclass=PoolMeta):
     __name__ = 'stock.shipment.in'
 
-    @classmethod
-    def _get_inventory_moves(cls, incoming_move):
-        move = super(ShipmentIn, cls)._get_inventory_moves(incoming_move)
+    def _get_inventory_move(self, incoming_move):
+        move = super()._get_inventory_move(incoming_move)
         if move:
-            move.set_product_location(
-                warehouse=incoming_move.shipment.warehouse.id)
+            move.set_product_location(warehouse=self.warehouse.id)
         return move
 
 
 class ShipmentOutReturn(metaclass=PoolMeta):
     __name__ = 'stock.shipment.out.return'
 
-    @classmethod
-    def _get_inventory_moves(cls, incoming_move):
-        move = super(ShipmentOutReturn, cls)._get_inventory_moves(
-            incoming_move)
+    def _get_inventory_move(self, incoming_move):
+        move = super()._get_inventory_move(incoming_move)
         if move:
-            move.set_product_location(
-                warehouse=incoming_move.shipment.warehouse.id)
+            move.set_product_location(warehouse=self.warehouse.id)
         return move
